import numpy as np

import autofit as af


class TestDataFit:

    def test__image_and_model_are_identical__no_masking__check_values_are_correct(self):
        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, False, False, False])

        model_data = np.array([1.0, 2.0, 3.0, 4.0])

        data_fit = af.fit.DataFit(data=data, noise_map=noise_map, mask=mask,
                                  model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, False, False, False])).all()
        assert (data_fit.model_data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.residual_map == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert (data_fit.chi_squared_map == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert data_fit.chi_squared == 0.0
        assert data_fit.reduced_chi_squared == 0.0
        assert data_fit.noise_normalization == np.sum(
            np.log(2 * np.pi * noise_map ** 2.0))
        assert data_fit.likelihood == -0.5 * (
                data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__no_masking__check_values_are_correct(self):
        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, False, False, False])

        model_data = np.array([1.0, 1.0, 1.0, 1.0])

        data_fit = af.fit.DataFit(data=data, noise_map=noise_map, mask=mask,
                                  model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, False, False, False])).all()
        assert (data_fit.model_data == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (data_fit.residual_map == np.array([0.0, 1.0, 2.0, 3.0])).all()
        assert (data_fit.chi_squared_map == np.array(
            [0.0, (1.0 / 2.0) ** 2.0, (2.0 / 2.0) ** 2.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squared == (1.0 / 2.0) ** 2.0 + (2.0 / 2.0) ** 2.0 + (
                3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squared == data_fit.chi_squared / 4.0
        assert data_fit.noise_normalization == np.sum(
            np.log(2 * np.pi * noise_map ** 2.0))
        assert data_fit.likelihood == -0.5 * (
                data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__include_masking__check_values_are_correct(self):
        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, True, True, False])

        model_data = np.array([1.0, 1.0, 1.0, 1.0])

        data_fit = af.fit.DataFit(data=data, noise_map=noise_map, mask=mask,
                                  model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, True, True, False])).all()
        assert (data_fit.model_data == model_data).all()
        assert (data_fit.residual_map == np.array([0.0, 0.0, 0.0, 3.0])).all()
        assert (data_fit.chi_squared_map == np.array(
            [0.0, 0.0, 0.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squared == (3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squared == data_fit.chi_squared / 2.0
        assert data_fit.noise_normalization == 2.0 * np.sum(
            np.log(2 * np.pi * 2.0 ** 2.0))
        assert data_fit.likelihood == -0.5 * (
                data_fit.chi_squared + data_fit.noise_normalization)


class TestDataFit1d:

    def test__image_and_model_mismatch__include_masking__check_values_are_correct(self):
        data_1d = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map_1d = np.array([2.0, 2.0, 2.0, 2.0])
        mask_1d = np.array([False, True, True, False])
        model_data_1d = np.array([1.0, 1.0, 1.0, 1.0])

        data_fit_1d = af.fit.DataFit1D(data_1d=data_1d, noise_map_1d=noise_map_1d,
                                       mask_1d=mask_1d,
                                       model_data_1d=model_data_1d)

        assert (data_fit_1d.data_1d == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit_1d.noise_map_1d == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit_1d.signal_to_noise_map_1d == np.array(
            [0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit_1d.mask_1d == np.array([False, True, True, False])).all()
        assert (data_fit_1d.model_data_1d == model_data_1d).all()
        assert (data_fit_1d.residual_map_1d == np.array([0.0, 0.0, 0.0, 3.0])).all()
        assert (data_fit_1d.chi_squared_map_1d == np.array(
            [0.0, 0.0, 0.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit_1d.chi_squared == (3.0 / 2.0) ** 2.0
        assert data_fit_1d.reduced_chi_squared == data_fit_1d.chi_squared / 2.0
        assert data_fit_1d.noise_normalization == 2.0 * np.sum(
            np.log(2 * np.pi * 2.0 ** 2.0))
        assert data_fit_1d.likelihood == -0.5 * (
                data_fit_1d.chi_squared + data_fit_1d.noise_normalization)



import os

import pytest

import autofit as af
import test.mock


@pytest.fixture(name="results")
def make_results_collection():
    results = af.ResultsCollection()

    results.add("first phase", "one")
    results.add("second phase", "two")

    return results


class TestResultsCollection(object):
    def test_with_name(self, results):
        assert results.from_phase("first phase") == "one"
        assert results.from_phase("second phase") == "two"

    def test_with_index(self, results):
        assert results[0] == "one"
        assert results[1] == "two"
        assert results.first == "one"
        assert results.last == "two"
        assert len(results) == 2

    def test_missing_result(self, results):
        with pytest.raises(af.exc.PipelineException):
            results.from_phase("third phase")


class MockPhase(af.AbstractPhase):
    def make_result(self, result, analysis):
        pass

    def __init__(self, phase_name, optimizer=None):
        super().__init__(phase_name)
        self.optimizer = optimizer or af.NonLinearOptimizer(phase_name)
        self.phase_path = phase_name
        self.phase_tag = phase_name

    def save_metadata(self, data_name, pipeline_name):
        pass


class MockHyperPhase(MockPhase, af.HyperPhase):
    pass


class TestPipeline(object):
    def test_hyper_phase_naming(self):
        pipeline = af.Pipeline("name", MockPhase("one"), MockHyperPhase("hyper"))

        # noinspection PyUnusedLocal,PyShadowingNames
        def func(phase, results):
            pass

        results = pipeline.run_function(func)

        assert len(results) == 2
        assert "hyper" not in results

    def test_hyper_phase_must_be_after_normal_phase(self):
        pipeline = af.Pipeline("name", MockHyperPhase("hyper"), MockPhase("one"))

        # noinspection PyUnusedLocal,PyShadowingNames
        def func(phase, results):
            pass

        with pytest.raises(af.exc.PipelineException):
            pipeline.run_function(func)

    def test_unique_phases(self):
        af.Pipeline("name", MockPhase("one"), MockPhase("two"))
        with pytest.raises(af.exc.PipelineException):
            af.Pipeline("name", MockPhase("one"), MockPhase("one"))

    def test_optimizer_assertion(self):
        optimizer = af.NonLinearOptimizer("Phase Name")
        optimizer.variable.profile = test.mock.GeometryProfile
        phase = MockPhase("phase_name", optimizer)

        try:
            os.makedirs(phase.make_path())
        except FileExistsError:
            pass

        phase.save_optimizer_for_phase()
        phase.assert_optimizer_pickle_matches_for_phase()

        optimizer.variable.profile.centre_0 = af.UniformPrior()

        with pytest.raises(af.exc.PipelineException):
            phase.assert_optimizer_pickle_matches_for_phase()

    def test_name_composition(self):
        first = af.Pipeline("first")
        second = af.Pipeline("second")

        assert (first + second).pipeline_name == "first + second"

    def test_assert_and_save_pickle(self):
        phase = af.AbstractPhase("name")

        phase.assert_and_save_pickle()
        phase.assert_and_save_pickle()

        phase.variable.galaxy = test.mock.Galaxy

        with pytest.raises(af.exc.PipelineException):
            phase.assert_and_save_pickle()


# noinspection PyUnresolvedReferences
class TestPhasePipelineName(object):
    def test_name_stamping(self):
        one = MockPhase("one")
        two = MockPhase("two")
        af.Pipeline("name", one, two)

        assert one.pipeline_name == "name"
        assert two.pipeline_name == "name"

    def test_no_restamping(self):
        one = MockPhase("one")
        two = MockPhase("two")
        pipeline_one = af.Pipeline("one", one)
        pipeline_two = af.Pipeline("two", two)

        composed_pipeline = pipeline_one + pipeline_two

        assert composed_pipeline[0].pipeline_name == "one"
        assert composed_pipeline[1].pipeline_name == "two"

        assert one.pipeline_name == "one"
        assert two.pipeline_name == "two"



from . import fit_util
from . import path_util
from . import text_util
from .dimension_type import DimensionType
from .dimension_type import map_types
from .dimension_type import map_types
from .fit import DataFit
from .fit import DataFit1D
from .phase import AbstractPhase
from .phase import as_grid_search
from .phase import HyperPhase
from .phase_property import PhaseProperty
from .pipeline import Pipeline
from .pipeline import ResultsCollection



import os
import shutil
import autofit as af

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

test_data_path = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__)))


class TestPhaseFoldersFromPipelineName:

    def test__phase_folders_is_none__returns_pipeline_name(self):
        phase_folders = af.path_util.phase_folders_from_phase_folders_and_pipeline_name(
            phase_folders=None,
            pipeline_name='test')

        assert phase_folders == ['test']

    def test__phase_folders_are_not_none__returns_phase_folders_and_pipeline_name(self):
        phase_folders = af.path_util.phase_folders_from_phase_folders_and_pipeline_name(
            phase_folders=['folder1', 'folder2'],
            pipeline_name='test')

        assert phase_folders == ['folder1', 'folder2', 'test']


class PathFromFolderNames:

    def test__1_directory_input__makes_directory__returns_path(self):
        path = af.path_util.path_from_folder_names(folder_names=['test1'])

        assert path == 'test1/'

    def test__multiple_directories_input__makes_directory_structure__returns_full_path(
            self):
        path = af.path_util.path_from_folder_names(
            folder_names=['test1', 'test2', 'test3'])

        assert path == 'test1/test2/test3/'


class TestMakeAndReturnPath:

    def test__1_directory_input__makes_directory__returns_path(self):
        path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=test_data_path, folder_names=['test1'])

        assert path == test_data_path + 'test1/'
        assert os.path.exists(path=test_data_path + 'test1')

        shutil.rmtree(test_data_path + 'test1')

    def test__multiple_directories_input__makes_directory_structure__returns_full_path(
            self):
        path = af.path_util.make_and_return_path_from_path_and_folder_names(
            path=test_data_path, folder_names=['test1', 'test2', 'test3'])

        assert path == test_data_path + 'test1/test2/test3/'
        assert os.path.exists(path=test_data_path + 'test1')
        assert os.path.exists(path=test_data_path + 'test1/test2')
        assert os.path.exists(path=test_data_path + 'test1/test2/test3')

        shutil.rmtree(test_data_path + 'test1')



import os

import pytest

import autofit as af
import autofit.mapper.prior_model.abstract
import autofit.mapper.prior_model.prior_model
import autofit.optimize.non_linear.non_linear
import test.mock

directory = os.path.dirname(os.path.realpath(__file__))

af.conf.instance = af.conf.Config("{}/../../workspace/config".format(directory),
                                  "{}/../../workspace/output/".format(directory))


class NLO(autofit.optimize.non_linear.non_linear.NonLinearOptimizer):
    def fit(self, analysis):
        class Fitness(object):
            def __init__(self, instance_from_physical_vector, constant):
                self.result = None
                self.instance_from_physical_vector = instance_from_physical_vector
                self.constant = constant

            def __call__(self, vector):
                instance = self.instance_from_physical_vector(vector)
                for key, value in self.constant.__dict__.items():
                    setattr(instance, key, value)

                likelihood = analysis.fit(instance)
                self.result = autofit.optimize.non_linear.non_linear.Result(instance,
                                                                            likelihood)

                # Return Chi squared
                return -2 * likelihood

        fitness_function = Fitness(self.variable.instance_from_physical_vector,
                                   self.constant)
        fitness_function(self.variable.prior_count * [0.5])

        return fitness_function.result


@pytest.fixture(name='phase')
def make_phase():
    class MyPhase(af.AbstractPhase):
        prop = af.PhaseProperty("prop")

    return MyPhase(phase_name='', optimizer_class=NLO)


@pytest.fixture(name='list_phase')
def make_list_phase():
    class MyPhase(af.AbstractPhase):
        prop = af.PhaseProperty("prop")

    return MyPhase(phase_name='', optimizer_class=NLO)


class TestPhasePropertyList(object):
    # def test_constants(self, list_phase):
    #     objects = [mock.Galaxy(), mock.Galaxy()]
    #
    #     list_phase.prop = objects
    #
    #     assert list_phase.constant.prop == objects
    #     assert list_phase.prop == objects

    def test_classes(self, list_phase):
        objects = [test.mock.GalaxyModel(), test.mock.GalaxyModel()]

        list_phase.prop = objects

        assert list_phase.variable.prop == objects
        assert list_phase.prop == objects

    def test_abstract_prior_models(self, list_phase):
        objects = [af.AbstractPriorModel(), af.AbstractPriorModel()]

        list_phase.prop = objects

        assert list_phase.variable.prop == objects
        assert list_phase.prop == objects

    def test_mix(self, list_phase):
        objects = [test.mock.GalaxyModel(), test.mock.Galaxy()]

        list_phase.prop = objects

        assert list_phase.variable.prop == objects
        assert list_phase.prop == objects

    def test_set_item(self, list_phase):
        galaxy_prior_0 = test.mock.GalaxyModel()
        objects = [galaxy_prior_0, test.mock.Galaxy()]

        list_phase.prop = objects

        galaxy_prior_1 = test.mock.GalaxyModel()
        list_phase.prop[1] = galaxy_prior_1

        assert list_phase.variable.prop == [galaxy_prior_0, galaxy_prior_1]

        galaxy = test.mock.Galaxy()

        list_phase.prop[0] = galaxy
        assert list_phase.prop == [galaxy, galaxy_prior_1]


class TestPhasePropertyCollectionAttributes(object):
    def test_set_list_as_dict(self, list_phase):
        galaxy_model = test.mock.GalaxyModel()
        list_phase.prop = dict(one=galaxy_model)

        assert len(list_phase.prop) == 1
        # noinspection PyUnresolvedReferences
        assert list_phase.prop.one == galaxy_model

    def test_override_property(self, list_phase):
        galaxy_model = test.mock.GalaxyModel()

        list_phase.prop = dict(one=test.mock.GalaxyModel())

        list_phase.prop.one = galaxy_model

        assert len(list_phase.prop) == 1
        assert list_phase.prop.one == galaxy_model

    def test_named_list_items(self, list_phase):
        galaxy_model = test.mock.GalaxyModel()
        list_phase.prop = [galaxy_model]

        # noinspection PyUnresolvedReferences
        assert getattr(list_phase.prop, "0") == galaxy_model

    def test_mix(self, list_phase):
        objects = dict(one=test.mock.GalaxyModel(), two=test.mock.Galaxy())

        list_phase.prop = objects

        list_phase.prop.one = test.mock.Galaxy()

        assert len(list_phase.variable.prop) == 2

    def test_named_attributes_in_variable(self, list_phase):
        galaxy_model = test.mock.GalaxyModel(variable_redshift=True)
        list_phase.prop = dict(one=galaxy_model)

        assert list_phase.variable.prior_count == 1
        assert list_phase.variable.prop.one == galaxy_model

        instance = list_phase.variable.instance_from_prior_medians()

        assert instance.prop.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_variable_override(self, list_phase):
        list_phase.prop = dict(one=test.mock.GalaxyModel())

        assert list_phase.variable.prior_count == 0

        galaxy_model = test.mock.GalaxyModel(variable_redshift=True)

        list_phase.prop.one = galaxy_model

        assert list_phase.variable.prior_count == 1
        # assert list_phase.variable.one == galaxy_model

        instance = list_phase.variable.instance_from_prior_medians()

        # assert instance.one is not None
        assert len(instance.prop) == 1

    def test_named_attributes_in_constant(self, list_phase):
        galaxy = test.mock.Galaxy()
        list_phase.prop = dict(one=galaxy)

        assert list_phase.variable.prior_count == 0
        assert list_phase.variable.prop.one == galaxy

    # def test_singular_model_info(self, list_phase):
    #     galaxy_model = test.mock.GalaxyModel(variable_redshift=True)
    #     list_phase.prop = dict(one=galaxy_model)
    #
    #     assert list_phase.variable.prop.one == galaxy_model
    #     assert len(galaxy_model.flat_prior_model_tuples) == 1
    #     assert len(galaxy_model.prior_tuples) == 1
    #
    #     assert len(list_phase.variable.flat_prior_model_tuples) == 1
    #
    #     print(list_phase.variable.info)
    #
    #     assert len(list_phase.variable.info.split('\n')) == 7

    def test_shared_priors(self, list_phase):
        list_phase.prop = dict(one=test.mock.GalaxyModel(variable_redshift=True),
                               two=test.mock.GalaxyModel(variable_redshift=True))

        assert list_phase.variable.prior_count == 2

        # noinspection PyUnresolvedReferences
        list_phase.prop.one.redshift = list_phase.prop.two.redshift

        assert list_phase.variable.prior_count == 1

    def test_hasattr(self, list_phase):
        list_phase.prop = dict()

        assert not hasattr(list_phase.prop, "one")
        list_phase.prop = dict(one=test.mock.GalaxyModel(variable_redshift=True))

        assert hasattr(list_phase.prop, "one")

    def test_position_not_a_prior(self, list_phase):
        list_phase.prop = [
            af.PriorModel(test.mock.Galaxy)]

        assert list_phase.variable.prior_count == 1
        assert "redshift" == list_phase.variable.prior_tuples_ordered_by_id[0][0]

        prior_model = af.PriorModel(test.mock.Galaxy)
        prior_model.phase_property_position = 0

        print(prior_model.constant_tuples)
        assert len(prior_model.constant_tuples) == 0



import copy

from .model_object import ModelObject


class AbstractModel(ModelObject):
    def __add__(self, other):
        instance = self.__class__()

        def add_items(item_dict):
            for key, value in item_dict.items():
                if isinstance(value, list) and hasattr(instance, key):
                    setattr(instance, key, getattr(instance, key) + value)
                else:
                    setattr(instance, key, value)

        add_items(self.__dict__)
        add_items(other.__dict__)
        return instance

    def copy(self):
        return copy.deepcopy(self)

    def object_for_path(self, path: (str,)) -> object:
        """
        Get the object at a given path.

        Parameters
        ----------
        path
            A tuple describing the path to an object in the model tree

        Returns
        -------
        object
            The object
        """
        instance = self
        for name in path:
            instance = getattr(instance, name)
        return instance

    def instances_of(self, cls: type) -> [object]:
        """
        Traverse the model tree returning all instances of the class

        Parameters
        ----------
        cls
            The type of objects to return

        Returns
        -------
        instances
            A list of instances of the type
        """
        return [
            instance for source in
            [
                list(self.__dict__.values())
            ] +
            [
                ls for ls in self.__dict__.values() if
                isinstance(
                    ls,
                    list
                )
            ] for
            instance in
            source if isinstance(
                instance,
                cls
            )
        ]

    def path_instance_tuples_for_class(self, cls: type, ignore_class=None):
        """
        Tuples containing the path tuple and instance for every instance of the class
        in the model tree.

        Parameters
        ----------
        ignore_class
            Children of instances of this class are ignored
        cls
            The type to find instances of

        Returns
        -------
        path_instance_tuples: [((str,), object)]
            Tuples containing the path to and instance of objects of the given type.
        """
        return path_instances_of_class(self, cls, ignore_class=ignore_class)

    def tuples_with_type(self, class_type):
        return list(filter(lambda t: t[0] != "id" and isinstance(t[1], class_type),
                           self.__dict__.items()))


def path_instances_of_class(obj, cls, ignore_class=None):
    if ignore_class is not None and isinstance(obj, ignore_class):
        return []
    if isinstance(obj, cls):
        return [(tuple(), obj)]
    results = []
    try:
        for key, value in obj.__dict__.items():
            for item in path_instances_of_class(value, cls, ignore_class=ignore_class):
                results.append(((key, *item[0]), item[1]))
        return results
    except AttributeError:
        return []


class ModelInstance(AbstractModel):
    """
    An object to hold model instances produced by providing arguments to a model mapper.

    @DynamicAttrs
    """

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __getitem__(self, item):
        return self.items[item]

    @property
    def items(self):
        return [value for key, value in self.__dict__.items() if
                key not in ("id", "component_number", "item_number")]

    def __len__(self):
        return len(self.items)
